#version 330 core
in vec2 TexCoord;
in vec3 Normal;
in vec3 FragPos;

struct Material
{
	sampler2D texture_diffuse1;
	sampler2D texture_diffuse2;
	sampler2D texture_specular1;
	sampler2D texture_specular2;
	float shininess;
};

struct DirLight
{
	vec3 direction;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};

struct PointLight
{
	vec3 position;
	float constant;
	float linear;
	float quadratic;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};

struct SpotLight
{
	vec3 position;
	vec3 direction;
	float constant;
	float linear;
	float quadratic;
	float innerCutoff;
	float outerCutoff;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};


uniform DirLight dirLight;
//#define NR_POINT_LIGHTS 3
//uniform PointLight pointLights[NR_POINT_LIGHTS];
//uniform SpotLight spotLight;

uniform vec3 viewPos;
//uniform vec3 light_position;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

uniform sampler2D tex_refraction;
uniform sampler2D tex_reflection;
uniform sampler2D tex_diffuse;

uniform float time;
uniform uint water_effect;
uniform float fresnel_base; // Base reflectivity for water
uniform bool inv_tex_reflection;

uniform vec4 clippingBounds; // (minX, maxX, minZ, maxZ)

out vec4 color;

// Function prototypes
float get_light_lum();

float gen_wave(float x, float y, float a, vec2 d, float w, float phi) {
    return a * sin(dot(normalize(d), vec2(x, y))*w + time * phi);
}

float get_wave(float x, float y) {
    float aptitude =  water_effect + 1.0;
    const float w = 16;
    float wave = gen_wave(x, y, 0.001, vec2(1, 0), w, 1.0) 
               + gen_wave(x, y, 0.001, vec2(1, -1), w, 1.0) 
               + gen_wave(x, y, 0.001, vec2(1, 1), w, 1.0);

    wave *= aptitude;
    return wave;
}

// Fresnel Schlick
float fresnelSchlick(float cosTheta, float F0) {
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}

void main(){

    // if (gl_FragCoord.x < 200.0 && gl_FragCoord.y < 200.0) {
    // color = vec4(abs(FragPos.x) / 10.0, abs(FragPos.z) / 10.0, 0.0, 1.0);
    // return;
    // } // TEST

    if (FragPos.x < clippingBounds.x || FragPos.x > clippingBounds.y || 
        FragPos.z < clippingBounds.z || FragPos.z > clippingBounds.w)
        discard;



    // Screen space position and wave distortion
    vec4 screen_pos = projection*view*vec4(FragPos, 1.0);
    float wave = get_wave(TexCoord.x, TexCoord.y);
    
    vec2 refraction = screen_pos.xy*0.5/screen_pos.w+vec2(0.5, 0.5) + vec2(wave, wave);
    vec2 reflection = vec2(refraction.x, 1-refraction.y) + vec2(wave, wave);
    if (inv_tex_reflection) {
		reflection.x = 1.0 - reflection.x;
		reflection.y = 1.0 - reflection.y;
	}
	
    // Sample textures
    vec3 colour_refraction = texture(tex_refraction, refraction).rgb;
    vec3 colour_reflection = texture(tex_reflection, reflection).rgb;
    
    vec3 direction_camera = normalize(viewPos-FragPos);
    //float angle_view = abs(dot(direction_camera, Normal));
    // final color = reflection + refraction + light_lum
    //color.rgb = (1 - angle_view) * colour_reflection + angle_view * colour_refraction + get_light_lum();
    //color.rgb = colour_reflection;
	
	
	float cosTheta = max(dot(direction_camera, Normal), 0.0);
    float fresnelReflectance = fresnelSchlick(cosTheta, fresnel_base);
	if (fresnel_base == 0.0) {
	    fresnelReflectance = 0.0;
    }
    vec3 mixedColor = fresnelReflectance * colour_reflection + (1.0 - fresnelReflectance) * colour_refraction;
    float lightLum = get_light_lum();
    color.rgb = mixedColor + vec3(lightLum);
	
	
    color.a = 1.0;
}


float get_light_lum() {
    vec3 x0 = vec3(TexCoord.x-0.01, get_wave(TexCoord.x-0.01, TexCoord.y), TexCoord.y);
    vec3 x1 = vec3(TexCoord.x+0.01, get_wave(TexCoord.x+0.01, TexCoord.y), TexCoord.y);
    vec3 y0 = vec3(TexCoord.x, get_wave(TexCoord.x, TexCoord.y-0.01), TexCoord.y-0.01);
    vec3 y1 = vec3(TexCoord.x, get_wave(TexCoord.x, TexCoord.y+0.01), TexCoord.y+0.01);

    vec3 normal_wave = normalize(cross(x1-x0, y1-y0));

    // todo light position
    //vec3 light_dir = normalize(light_position-FragPos);
    vec3 light_dir = normalize(-dirLight.direction);
    float diffuse_light = dot(normal_wave, light_dir);
    //float light_dist = length(light_position-FragPos);
    float light_dist = 2.0f;
    diffuse_light /= 1+pow(light_dist, -0.5);

    vec3 specular = 2*normal_wave*dot(normal_wave, light_dir)-light_dir;
    specular = normalize(specular);
    vec3 view_dir = normalize(viewPos-FragPos);

    float spec_light = pow(max(dot(specular, view_dir), 0.0), 64);
    spec_light = clamp(spec_light, 0.0, 1.0);

    float lum = 0.5*diffuse_light+spec_light;
    return lum;
}